# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division, print_function

import pytz

from django.utils.six.moves.urllib.parse import urlparse, urlunparse, parse_qs,\
                                                urlencode
import json
from datetime import timedelta

from django.db import models, transaction, IntegrityError
from django.utils import timezone
from django.utils.html import mark_safe
from django.urls import reverse
from django.core.validators import URLValidator
from django.core.exceptions import ValidationError
from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from django.utils.encoding import python_2_unicode_compatible

from commonstuff.models_base import ModelWith2Images

from .settings import DEFAULT_ICON_URL, ERROR_THRESHOLD, FCM_URL


# http://stackoverflow.com/questions/27072222/django-1-7-1-makemigrations-fails-when-using-lambda-as-default-for-attribute
def _timezone_now():
    return timezone.now()


class BaseSubscription(models.Model):
    """Base class for push subscriptions"""
    
    endpoint = models.CharField(_('endpoint'), max_length=255, unique=True, editable=False)
    key = models.CharField(_('key'), max_length=255, blank=True, default='', editable=False)
    auth_secret = models.CharField(_('auth secret'), max_length=255, blank=True, default='', editable=False)
    is_active = models.BooleanField(_('is active'), default=True, db_index=True)
    errors = models.PositiveIntegerField(_('send errors'), default=0, editable=False)
    ua = models.CharField(_('useragent'), max_length=255, editable=False)
    timezone = models.CharField(
        _('timezone'), max_length=48,
        choices=((x,x) for x in sorted(pytz.all_timezones_set)),
        db_index=True
    )
    created_at = models.DateTimeField(_('created'), default=_timezone_now, editable=False)
    activated_at = models.DateTimeField(_('activated'), blank=True, null=True, editable=False)
    deactivated_at = models.DateTimeField(_('deactivated'), blank=True, null=True, editable=False)
    
    class Meta:
        abstract = True
    
    def save(self, *args, **kwargs):
        if self.is_active and self.activated_at is None:
            self.activated_at = timezone.now()
        return super(BaseSubscription, self).save(*args, **kwargs)
    
    @classmethod
    def count_active(cls):
        return cls.objects.filter(is_active=True).count()
    
    def endpoint_truncated(self):
        return "%s..." % self.endpoint[0:64]
    endpoint_truncated.admin_order_field = 'endpoint'
    endpoint_truncated.short_description = _('endpoint')

    def is_gcm(self):
        """Is it a Google subscription (they have own sending algo for now on)"""
        return self.is_fcm()
    
    def is_fcm(self):
        return self.endpoint.startswith(FCM_URL)
    
    def supports_payload(self):
        return bool(self.key and self.auth_secret)
    
    def endpoint_and_keys(self):
        """
        Subscription info for payload encryption in a format of
        https://developer.mozilla.org/ru/docs/Web/API/PushSubscription/toJSON
        """
        d = { "endpoint": self.endpoint, }
        if self.supports_payload():
            d['keys'] = {
                "auth": self.auth_secret,
                "p256dh": self.key,
            }
        return d
    
    def deactivate(self):
        if self.is_active:
            self.is_active = False
            self.deactivated_at = timezone.now()
        return self
    
    def reactivate_if_needed(self):
        if self.id:
            self.errors = 0
            if not self.is_active:
                self.is_active = True
                self.activated_at = timezone.now()
        return self
    
    def errors_accounting(self, quantity=1):
        if quantity > 0:
            self.errors += quantity
            if self.errors >= ERROR_THRESHOLD:
                self.deactivate()
        elif quantity < 0:
            # subscription had errors, but it's ok now, so decrease error counter
            if self.errors > 0:
                self.errors += quantity  # negative number here
                if self.errors < 0:
                    self.errors = 0
        return self
    
    def push_service_response_to_errors(self, response):
        """
        Parse response from remote push-server.
        
        Penalize erroneous subscriptions, removes points from successful delivery.
        Receives the response body of the push server (json str or dict).
        """
        if response and not isinstance(response, dict):
            response = json.loads(response)
        
        if self.is_fcm():
            return self.__fcm_push_service_response_to_errors(response)
        else:
            return self.__normal_push_service_response_to_errors(response)
    
    def __normal_push_service_response_to_errors(self, response):
        # For non(!)-success responses, an extended error code object will be returned
        if response:  # we have an error
            if response['code'] < 300:
                return self
            # http://autopush.readthedocs.io/en/latest/http.html#response
            error_points = 0 # default for 401, 413, 500, 503
            if response['code'] == 301:
                error_points = 3
            elif response['code'] == 400 and response['errno'] < 111:
                error_points = 1
            elif response['code'] in (404, 410):
                error_points = 15
            if error_points:
                self.errors_accounting(error_points).save()
        else:  # ok
            # check so we don't query database for no reason
            if self.errors > 0:
                self.errors_accounting(-1).save()
        return self
    
    def __fcm_push_service_response_to_errors(self, response):
        response_result = response['results'][0]  # part we need
        # ok, decrease error counter
        if 'message_id' in response_result:
            if self.errors > 0:
                self.errors_accounting(-1).save()
            # If registration_id is set, replace the original ID with the new value
            # (canonical ID) in your server database.
            if 'registration_id' in response_result:
                new_endpoint = response_result['registration_id']
                try:
                    # it's not clear from the FCM docs if they respond
                    # with the full URL or only a endpoint part
                    url_validate = URLValidator(schemes=['https',])
                    url_validate(new_endpoint)
                except ValidationError:
                    new_endpoint = "%s/%s" % (FCM_URL, new_endpoint)
                self.endpoint = new_endpoint
                try:
                    self.save()
                except IntegrityError as e:
                    # Duplicate entry for endpoint - new canonical endpoint
                    # is already stored in the DB 
                    pass
        elif 'error' in response_result:
            error_code = response_result['error']
            # https://firebase.google.com/docs/cloud-messaging/http-server-ref#error-codes
            if error_code in ('NotRegistered', 'InvalidRegistration',):
                # google docs say we should turn off these subscriptions immediately
                # but lets give them one more chance (based on testing responses manually)
                self.errors_accounting(+15).save()
            # The server couldn't process the request in time. Retry the same request
            elif error_code in ('Unavailable', 'InternalServerError',):
                pass
            # misc errors
            else:
                self.errors_accounting(+1).save()
        return self


@python_2_unicode_compatible
class DigestSubscription(BaseSubscription):
    """Digest push subscription (the only actual subscription type for now)"""
    
    class Meta:
        verbose_name = _('push subscription')
        verbose_name_plural = _('push subscriptions')
    
    def __str__(self):
        return _('subscription ID %(pk)d') % { 'pk': self.id }


class PublicTaskManager(models.Manager):
    """Push tasks, that are available for subscribers"""
    def get_queryset(self):
        return super(PublicTaskManager, self).get_queryset() \
                    .filter(is_active=True) \
                    .exclude(started_at__isnull=True)


@python_2_unicode_compatible
class Task(ModelWith2Images):
    """Push tasks (the task here is to send something to push subscribers)"""
    
    title = models.CharField(_('title'), max_length=61)
    message = models.TextField(_('message'), max_length=122)
    url = models.URLField(
        _('url'), max_length=512,
        help_text=_('If there is no from param in url, ?from=push will be added automatically.')
    )
    is_active = models.BooleanField(
        _('active'), default=False, db_index=True,
        help_text=_('Is it an active task that should be done, or a draft.')
    )
    views = models.PositiveIntegerField(_('views'), default=0, editable=False)
    clicks = models.PositiveIntegerField(_('clicks'), default=0, editable=False)
    closings = models.PositiveIntegerField(
        _('closings'), default=0, editable=False,
        help_text=_('When user closes the notification.')
    )
    created_at = models.DateTimeField(_('created'), default=_timezone_now, editable=False)
    run_at = models.DateTimeField(
        _('run at'), default=_timezone_now, db_index=True,
        help_text=_('Send time here is set according to subscriber\'s timezone! Also you can postpone sending.')
    )
    started_at = models.DateTimeField(
        _('started at'), blank=True, null=True, db_index=True, editable=False,
        help_text=_('When we started to send this task.')
    )
    done_at = models.DateTimeField(
        _('done at'), blank=True, null=True, db_index=True, editable=False,
        help_text=_('When work on this task has been done.')
    )
    
    objects = models.Manager()  # default: all rows
    public_objects = PublicTaskManager()
    
    class Meta:
        verbose_name = _('push task')
        verbose_name_plural = _('push tasks')
    
    def __str__(self):
        return '%(verbose_name)s "%(current_title)s"' % {
            'verbose_name': self._meta.verbose_name,
            'current_title': self.title,
        }
    
    def get_absolute_url(self):
        return reverse('push_show_notification', args=[self.id,])
    
    def is_done(self):
        return self.done_at is not None
    is_done.admin_order_field = 'done_at'
    is_done.boolean = True
    is_done.short_description = _('done')
    
    def ctr(self, raw=False):
        """For admin interface"""
        try:
            ctr = self.clicks * 100.0 / self.views
            return ctr if raw else "%0.2f%%" % ctr
        except ZeroDivisionError:
            return None if raw else 'n/a'
    ctr.short_description = _('CTR')
    
    def closings_percent(self, raw=False):
        """For admin: rejections to views task ratio"""
        try:
            dtr = self.closings * 100.0 / self.views
            if raw:
                return dtr
            else:
                txt = _('%(num_closings)d times users rejected this notification') \
                    % { 'num_closings': self.closings }
                return mark_safe('<span title="%s">%0.2f%%</span>' % (txt, dtr))
        except ZeroDivisionError:
            return None if raw else 'n/a'
    closings_percent.short_description = _('rejections')
    
    def run_for(self, with_microseconds=False):
        """How much time we spend sending this task (textual representation for admin)"""
        if self.started_at is None:
            return None
        td = timedelta()
        for tzl in self.timezonelayout_set.all():
            td = td + tzl.run_for()
        if not with_microseconds:
            td = td - timedelta(microseconds=td.microseconds)
        return td if td != timedelta() else None
    run_for.short_description = _('run for')

    def url_relative(self, add_querystring={}):
        """Relative push-task url without domain for service worker"""
        # break url to components
        parts = urlparse(self.url)
        # add querystring if needed
        if add_querystring:
            qs = parse_qs(parts[4])
            for k in qs:
                # The dictionary keys are the unique query variable names
                # and the values are lists of values for each name
                qs[k] = qs[k][0]
            add_querystring.update(qs)
            parts = parts[0:4] + (urlencode(add_querystring),) + parts[5:]
        # join url from parts ignoring host and port
        return urlunparse(('', '') + parts[2:])
    
    def get_payload(self):
        """Task info we send to push-subscribers"""
        payload = {
            'title': self.title,
            'message': self.message,
            'icon': self.image.url if self.has_image() else DEFAULT_ICON_URL,
            'tag': 'notification-digest',
            'url': self.get_absolute_url(),
            'views_stat_url': reverse('push_notification_plus_one', args=['views', self.id,]),
            'closings_stat_url': reverse('push_notification_plus_one', args=['closings', self.id,]),
        }
        if self.has_image2():
            payload['image'] = self.image2.url
        return payload
    
    def all_timezones_done(self):
        return not self.timezonelayout_set.filter(done_at__isnull=True).exists()
    
    def save(self, *args, **kwargs):
        ret = super(Task, self).save(*args, **kwargs)
        
        # sending time can be edited only for tasks that are not started yet
        if self.started_at is None:
            # we check that task start time has changed by checking only
            # one timezone (otherwise admin push interface will be very slow)
            if len(pytz.all_timezones_set) == self.timezonelayout_set.count():
                test_layout_tz = self.timezonelayout_set.get(timezone=settings.TIME_ZONE)
                # empty timedelta means these datetimes are equal
                if self.run_at - test_layout_tz.run_at == timedelta():
                    return ret
            
            # detect hour only(!) for project's default timezone
            default_tz = pytz.timezone(settings.TIME_ZONE)  # project's timezone
            run_at_in_default_tz = default_tz.normalize( self.run_at.astimezone(default_tz) )
            # hour int for project's timezone (not UTC)
            # like the time editor sees in push admin interface
            # e.g. 10 for 07:30:00 UTC == 10:30:00 Europe/Moscow
            run_at_hour = int(run_at_in_default_tz.hour)
            
            # update layout db here
            with transaction.atomic():
                self.timezonelayout_set.all().delete()
                for tz_str in pytz.all_timezones_set:
                    tz = pytz.timezone(tz_str)
                    tzed_run_at = tz.normalize( self.run_at.astimezone(tz) )
                    # e.g. for Asia/Ekaterinburg it will be 10:30:00 Asia/Ekaterinburg
                    tzed_run_at = tzed_run_at.replace(hour=run_at_hour)
                    self.timezonelayout_set.create(
                        timezone=tz_str,
                        # Django ORM will change it to 05:30:00 UTC then saving
                        # to db that does not support timezones (mysql)
                        # TODO: wtf will happen for DB's that support timezones ???
                        run_at=tzed_run_at
                    )
        return ret


class UndoneTZLManager(models.Manager):
    """Timezone sub-tasks that have to be done"""
    def get_queryset(self):
        return super(UndoneTZLManager, self).get_queryset().filter(
            task__is_active=True,
            run_at__lte=timezone.now(),
            done_at__isnull=True,
            started_at__isnull=True
        ).order_by('run_at')


class PublicTZLManager(models.Manager):
    """Push sub-tasks, that are already available for subscribers"""
    def get_queryset(self):
        return super(PublicTZLManager, self).get_queryset() \
                    .filter(task__is_active=True) \
                    .exclude(started_at__isnull=True) \
                    .order_by('-run_at')


@python_2_unicode_compatible
class TimezoneLayout(models.Model):
    """
    Push task sub-tasks by timezone
    (so that users get "pushed" for the time in their local timezone).
    """
    task = models.ForeignKey(Task, editable=False, on_delete=models.CASCADE)
    timezone = models.CharField(
        _('timezone'), max_length=48,
        choices=((x,x) for x in sorted(pytz.all_timezones_set)),
        db_index=True, editable=False
    )
    run_at = models.DateTimeField(_('run at'), default=_timezone_now, db_index=True, editable=False)
    started_at = models.DateTimeField(_('started at'), blank=True, null=True, db_index=True, editable=False)
    done_at = models.DateTimeField(_('done at'), blank=True, null=True, db_index=True, editable=False)
    
    objects = models.Manager()
    undone_objects = UndoneTZLManager()
    public_objects = PublicTZLManager()
    
    class Meta:
        unique_together = ('task', 'timezone',)
    
    def __str__(self):
        return _('%(timezone)s tz sub-task for %(task)s') % {
            'timezone': self.timezone,
            'task': self.task
        }
    
    def run_for(self):
        if self.done_at is None or self.started_at is None:
            return timedelta()  # zero timedelta
        return self.done_at - self.started_at



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division, print_function

from pywebpush import WebPusher
import urllib3
from requests import exceptions as requests_ex

import time
import logging
import os
from multiprocessing import Pool
import json
from random import randint
from datetime import timedelta

from django.core.management.base import BaseCommand
from django.utils import timezone
from django.conf import settings

from commonstuff.models import PidLock

from push.models import DigestSubscription, TimezoneLayout, Task
from push.settings import FCM_SERVER_KEY, PUSHSEND_WORKERS


# setup custom logging if we have a folder for logs
logger = logging.getLogger(__name__)
log_dir = os.path.join(settings.BASE_DIR, 'log')
if os.access(log_dir, os.F_OK) and os.access(log_dir, os.W_OK):
    logger.propagate = False
    logger.setLevel(logging.INFO)
    logger.addHandler(logging.FileHandler(os.path.join(log_dir, 'pushsend.log')))
    if settings.DEBUG:
        db_logger = logging.getLogger('django.db.backends') # log all SQL's
        db_logger.propagate = False
        db_logger.setLevel(logging.DEBUG)
        db_logger.addHandler(logging.FileHandler(
            os.path.join(log_dir, 'pushsend_sql_debug.log')
        ))
else:
    logger.addHandler(logging.NullHandler())


# функция должна быть вынесена из класса, чтобы запускаться в Pool
# параметр она получает только один (в нашем случае tuple со всеми данными)
def send_push_worker(data):
    (
        subscr_list,
        payload,
        ttl,
    ) = data
    
    responses = []
    exceptions = []  # logger нельзя передать в worker
    for subscr in subscr_list:
        try:
            # http://autopush.readthedocs.io/en/latest/http.html#error-codes
            response = WebPusher( subscr.endpoint_and_keys() ).send(
                data=payload if subscr.supports_payload() else None,
                ttl=ttl,
                timeout=2,
                gcm_key=FCM_SERVER_KEY if subscr.is_fcm() else None
            )
            responses.append( (subscr, response) )
        except Exception as e:
            exceptions.append( (subscr, e, time.time(),) )
    
    return (responses, exceptions)


class Command(BaseCommand):
    """Рассылает push-уведомления по крону"""
    
    help = 'Pushes notification to web-subscribers (cron use only)'
    
    DB_LIMIT = 7000  # чтобы не сожрало память, подписчики из базы выбираются по X штук
    # Время жизни сообщения (int секунд).
    # None - поле вообще не будет задаваться нами для сообщения.
    # 0 - для Mozilla это, кажется, сразу убирать уведомление после получения.
    # FCM утверждает, что максимальное время у них 4 недели.
    # для VAPID 86400 секунд - это макс по стандарту
    TTL = 86400 - 1  # Пока так (больше и не нужно по факту)

    start_time = None
    pid_lock = None
    
    def __init__(self, *args, **kwargs):
        # protection against several copies of pushsend running at the same time
        self.pid_lock = PidLock(process=__file__)
        self.pid_lock.save_or_die()
        
        # все что после save_or_die может не выполниться,
        # если другой процесс уже выполняет команду
        
        self.start_time = time.time()
        logger.info("pushsend started at %s.", time.asctime(time.localtime(self.start_time)))
        super(Command, self).__init__(*args, **kwargs)
    
    def __del__(self):
        # может быть None, если это дублирующий процесс, который будет
        # убит в вызове save_or_die из конструктора
        if self.start_time is not None:
            end_time = time.time()
            logger.info("pushsend ended at %s.", time.asctime(time.localtime(end_time)))
            logger.info("pushsend worked for %f seconds.", (end_time - self.start_time))
        
        if self.pid_lock and self.pid_lock.pk:
            self.pid_lock.delete()

    def handle(self, *args, **options):
        urllib3.disable_warnings()
        self.clean_push_db()
        
        tz_layouts = list(TimezoneLayout.undone_objects.all())
        for tz_layout in tz_layouts:
            task = tz_layout.task
            
            tz_layout.started_at = timezone.now()
            tz_layout.save(update_fields=['started_at'])
            if not task.started_at:
                task.started_at = timezone.now()
                task.save(update_fields=['started_at'])
            
            # subscribers
            # Не можем отфильтровать здесь по is_active=True, т.к. сразу после
            # отправки в случае определенных ошибок выключаем часть подписок.
            # Это сдвигает срез выборки по active и тогда у пограничных между
            # пачками будет двойная рассылка. Поэтому отфильтруем активные перед
            # отправкой средствами питона, а не БД (вообще надо будет переделать).
            qs = DigestSubscription.objects \
                    .filter(timezone=tz_layout.timezone) \
                    .order_by('id')
            
            # send
            max_i = qs.count()
            # выбираем пачками из БД, чтобы объекты не пожрали память
            for i in range(0, max_i, self.DB_LIMIT):
                subscriptions = qs[i:(i+self.DB_LIMIT)]
                active_subscriptions = [ s for s in subscriptions if s.is_active ]
                subscriptions = None  # освободить память
                if not active_subscriptions:
                    continue  # next pack if no active subscribers in this one
                
                # разбиваем на кол-во частей или по кол-ву подписок, если их меньше чем ядер
                max_workers = min(PUSHSEND_WORKERS, len(active_subscriptions))
                per_worker = [ active_subscriptions[i::max_workers] for i in range(max_workers) ]
                active_subscriptions = None  # освобождаем память
                
                pool_data = []
                for subscr_chunk in per_worker:
                    pool_data.append(
                        (
                            # различаются
                            subscr_chunk,
                            # общие для всех вызовов
                            json.dumps( task.get_payload() ),
                            self.TTL,
                        )
                    )
                per_worker = None  # освобождаем память
                
                if max_workers == 1:
                    responses, exceptions = send_push_worker(pool_data[0])
                else:
                    pool = Pool(processes=max_workers)
                    multi_result = pool.map(send_push_worker, pool_data)
                    # https://stackoverflow.com/questions/20914828/python-multiprocessing-pool-join-not-waiting-to-go-on
                    # https://stackoverflow.com/questions/25391025/what-exactly-is-python-multiprocessing-modules-join-method-doing
                    pool.close()
                    pool.join()
                    responses = []
                    exceptions = []
                    # flattern result list
                    for result in multi_result:
                        responses += result[0]
                        exceptions += result[1]
                    multi_result = None
                pool_data = None
                
                # logger нельзя передавать в воркеры, поэтому залогим
                # полученные exceptions сейчас
                for subscr, e, timestamp in exceptions:
                    localtime = time.asctime(time.localtime(timestamp))
                    logger.error("%s, %s: %s, %s" % (__name__, localtime, type(e), e))
                    # endpoint is not url
                    if isinstance(e, requests_ex.InvalidURL) \
                    or isinstance(e, requests_ex.URLRequired): 
                        subscr.deactivate().save()
                
                # с изменениями БД в воркерах тоже непонятно как все работает
                for subscr, response in responses:
                    try:
                        subscr.push_service_response_to_errors(response.text)
                    except Exception as e:
                        localtime = time.asctime( time.localtime(time.time()) )
                        logger.exception("%s, %s: %s, %s" % (__name__, localtime, type(e), e))
            
            # запишем результаты наших стараний в БД
            tz_layout.done_at = timezone.now()
            tz_layout.save(update_fields=['done_at'])
            if task.all_timezones_done():
                task.done_at = timezone.now()
                task.save(update_fields=['done_at'])
    
    def clean_push_db(self):
        """
        Чистка старых tzl и неактивных подписок, чтобы таблица не разросталась.
        """
        if randint(0, 299):
            return
        
        # Старые tzl занимают место и раздувают таблицу (по которой активно
        # идет поиск во вьюхе last_notification). Единственная инфа, которую мы
        # берем оттуда это время, потраченное на рассылку, но она нужна только
        # для актуальных заданий и не имеет ценности для старья.
        tasks = Task.public_objects.filter(done_at__lt=( timezone.now() - timedelta(days=90) ))
        for task in tasks:
            task.timezonelayout_set.all().delete()
        
        # т.к. мы перебираем все подписки, а не только активные, то старые записи
        # скорее всего серьезно замедляют рассылку (но все не удаляем, чтобы можно
        # было собрать статистику при необходимости)
        DigestSubscription.objects.filter(
            is_active=False,
            deactivated_at__lt=( timezone.now() - timedelta(days=365) )
        ).delete()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division, print_function

import pytz
from PIL import Image

import logging

from django.db import IntegrityError, connection
from django.http import JsonResponse, Http404, HttpResponse
from django.views.decorators.cache import never_cache, cache_page
from django.views.decorators.http import require_POST
from django.views.decorators.csrf import csrf_exempt, csrf_protect
from django.shortcuts import render, redirect
from django.conf import settings
from django.urls import reverse, NoReverseMatch
from django.core.validators import URLValidator
from django.core.exceptions import ValidationError
from django.contrib.staticfiles import finders
from django.contrib.sites.shortcuts import get_current_site

from .settings import FCM_SENDER_ID, FCM_URL, APP_ICON_URLS, USE_CSRF, \
                      APP_BACKGROUND_COLOR, APP_THEME_COLOR
from .models import DigestSubscription, Task, TimezoneLayout


logger = logging.getLogger(__name__)


@never_cache
def on_off(request):
    return render(request, 'push/on_off.html')


def info_disable(request):
    return render(request, 'push/info_disable.html')


@cache_page(86400)
def manifest_json(request):
    # нам этот файл нужен, чтобы настроить gcm_sender_id
    # https://developers.google.com/web/updates/2014/11/Support-for-installable-web-apps-with-webapp-manifest-in-chrome-38-for-Android
    try:
        start_url = "%s" % reverse('homepage')
    except NoReverseMatch:
        start_url = "/"
    
    site = get_current_site(request)
    manifest = {
        "name": site.name,
        "short_name": site.name,
        "icons": [],
        "start_url": "%s?from=manifest_json" % start_url,
        "display": "standalone",
        "gcm_sender_id": FCM_SENDER_ID,
    }
    
    for icon in APP_ICON_URLS:
        # кеширование на сутки, так что тут все ок
        img_file = finders.find(icon)
        with Image.open(img_file) as img:
            manifest['icons'].append({
                "src": settings.STATIC_URL + icon,
                "sizes": "%dx%d" % (img.width, img.height),
                "type": "image/%s" % img.format.lower()
            })
    
    if APP_BACKGROUND_COLOR is not None:
        manifest['background_color'] = APP_BACKGROUND_COLOR
    if APP_THEME_COLOR is not None:
        manifest['theme_color'] = APP_THEME_COLOR
    
    return JsonResponse(manifest)


def _save(request):
    try:
        ua = request.META['HTTP_USER_AGENT'][:255]
    except (AttributeError, KeyError):
        ua = ''
    
    endpoint = request.POST.get('endpoint')
    logger.debug('Push save endpoint: %s' % endpoint)
    try:
        url_validate = URLValidator(schemes=['https',])
        url_validate(endpoint)
    except ValidationError:
        # chrome 44 и, кажется, 45 иногда передают только id без урла, костылим
        if endpoint and ua and 'Chrome' in ua:
            endpoint = "%s/%s" % (FCM_URL, endpoint)
            logger.debug('Fixed endpoint to: %s' % endpoint)
        else:
            # endpoint должен быть урлом или нах вообще всё это
            raise Http404('Wrong endpoint')
    
    # если юзер самостоятельно через браузер отпишется и подпишется потом,
    # у него изменится идентификатор подписки
    key = request.POST.get('key', '')
    auth_secret = request.POST.get('auth_secret', '')
    
    timezone = request.POST.get('timezone')
    if timezone not in pytz.all_timezones:
        timezone = settings.TIME_ZONE
    
    # поиск
    try:
        subscr = DigestSubscription.objects.get(endpoint=endpoint)
    except DigestSubscription.DoesNotExist:
        subscr = DigestSubscription()
        subscr.endpoint = endpoint
    # актуализация инфы
    subscr.key = key
    subscr.auth_secret = auth_secret
    subscr.timezone = timezone
    subscr.ua = ua
    subscr.reactivate_if_needed()
    subscr.save()
    
    return JsonResponse({
        'response': {
            'status': 'ok',
            'id': subscr.pk,
        }
    })


@require_POST
@csrf_exempt
def save(request):
    # https://docs.djangoproject.com/en/1.8/ref/csrf/#view-needs-protection-for-one-path
    @csrf_protect
    def _protected_save(request):
        return _save(request)
    
    if USE_CSRF:
        return _protected_save(request)
    else:
        return _save(request)


def _deactivate(request):
    endpoint = request.POST.get('endpoint')
    
    try:
        subscr = DigestSubscription.objects.get(endpoint=endpoint)
    except DigestSubscription.DoesNotExist:
        raise Http404
    subscr.deactivate().save()
    
    return JsonResponse({
        'response': {
            'status': 'ok',
            'id': subscr.pk,
        }
    })


@require_POST
@csrf_exempt
def deactivate(request):
    # https://docs.djangoproject.com/en/1.8/ref/csrf/#view-needs-protection-for-one-path
    @csrf_protect
    def _protected_deactivate(request):
        return _deactivate(request)
    
    if USE_CSRF:
        return _protected_deactivate(request)
    else:
        return _deactivate(request)


def _notification_plus_one(what, id):
    """
    Облегченная фактическая функция-учетчик статы (на живом SQL для скорости).
    Вынесена, чтобы можно было вызвать напрямую, не передавая request.
    """
    cursor = connection.cursor()
    cursor.execute("UPDATE %s SET %s=%s+1 WHERE id=%d" \
                    % (Task._meta.db_table, what, what, int(id)))


@never_cache
def notification_plus_one(request, what, id):
    """
    Вьюха учета просмотров|закрытий push-задания для статистики.
    """
    _notification_plus_one(what, id)
    return HttpResponse(content='ok')


@never_cache
def last_notification(request):
    """
    Возвращает данные о последнем актуальном задании
    (далее - актуальном для дефолтового часового пояса)
    для старых legacy-подписок, где нельзя шифровать данные.
    """
    try:
        # Выбираем последнее выполненное задание для дефолтового часового пояса.
        # Это старый обработчик для легаси подписок, которым мы не можем шифровать
        # данные, тк не знаем auth_secret, а раз не обновили auth_secret, то и
        # правильный часовой пояс у них не знаем (~97.3% случаев). Еще ~2.7%
        # случаев это новые подписки на некро-браузеры, которые имеют свой часовой
        # пояс, но не поддерживают шифрование - пох, такие будут отмирать и % падать.
        tz = TimezoneLayout.public_objects.filter(timezone=settings.TIME_ZONE)[0]
    except IndexError:
        raise Http404
    _notification_plus_one('views', tz.task.id)
    return JsonResponse({ 'notification': tz.task.get_payload(), })


@never_cache
def show_notification(request, id):
    try:
        task = Task.public_objects.get(pk=id)
    except Task.DoesNotExist:
        raise Http404
    _notification_plus_one('clicks', task.id)
    return redirect(task.url_relative(add_querystring={'from': 'push'}))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division, print_function

import re
from django.utils.six.moves import urllib
from datetime import timedelta

from PIL import Image

from django import forms
from django.conf import settings
from django.contrib.sites.models import Site
from django.template.defaultfilters import filesizeformat
from django.utils.translation import ugettext as _

from .models import Task
from .settings import MIN_ICON_H, MIN_ICON_W, ICON_CUSTOM_MAX_FILESIZE, \
                      MIN_BIG_IMAGE_W, MIN_BIG_IMAGE_H, BIG_IMAGE_MAX_FILESIZE


class TaskAdminForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super(TaskAdminForm, self).__init__(*args, **kwargs)
        
        ratio = TaskAdminForm._ratio_from_wh(MIN_ICON_W, MIN_ICON_H)
        self.fields['image'].help_text += _("""
            Push icon.
            Minimum %(min_w)dx%(min_h)d px,
            %(ratio_w)d:%(ratio_h)d aspect ratio.
            Optional (default icon will be used).
            """) % {
                'min_w': MIN_ICON_W,
                'min_h': MIN_ICON_H,
                'ratio_w': ratio[0],
                'ratio_h': ratio[1]
            }
        
        ratio = TaskAdminForm._ratio_from_wh(MIN_BIG_IMAGE_W, MIN_BIG_IMAGE_H)
        self.fields['image2'].help_text += _("""
            Big image for push in Chrome.
            Minimum %(min_w)dx%(min_h)d px,
            %(ratio_w)d:%(ratio_h)d aspect ratio.
            Optional (if so, there will be no big image).
            """) % {
                'min_w': MIN_BIG_IMAGE_W,
                'min_h': MIN_BIG_IMAGE_H,
                'ratio_w': ratio[0],
                'ratio_h': ratio[1]
            }
    
    class Meta:
        model = Task
        fields = ['title', 'message', 'url', 'is_active', 'image', 'image2', 'run_at',]
    
    def clean_image(self):
        image = self.cleaned_data.get('image')
        self._check_image(
            image,
            max_size=ICON_CUSTOM_MAX_FILESIZE*1024,  # in settings we use kb
            min_w=MIN_ICON_W,
            min_h=MIN_ICON_H
        )
        return image
    
    def clean_image2(self):
        image2 = self.cleaned_data.get('image2')
        self._check_image(
            image2,
            max_size=BIG_IMAGE_MAX_FILESIZE*1024,
            min_w=MIN_BIG_IMAGE_W,
            min_h=MIN_BIG_IMAGE_H
        )
        return image2
    
    @staticmethod
    def _ratio_from_wh(w, h):
        """Calculate aspect ratio from image w h"""
        a = w
        b = h
        while a != 0 and b != 0:
            if a > b:
                a %= b
            else:
                b %= a
        gcd = a + b
        return [w/gcd, h/gcd]
    
    def _check_image(self, image, max_size=None, min_w=None, min_h=None):
        if image:
            if max_size is not None:
                if hasattr(image, '_size') and image._size > max_size:
                    raise forms.ValidationError(
                        _("File size can't exceed %(max_size)s. This image is %(img_size)s.") % {
                            'max_size': filesizeformat(max_size),
                            'img_size': filesizeformat(image._size)
                        }
                    )
            
            img_pillow = Image.open(image)
            
            m = re.search(r'.+?\.([^.]+)$', str(image))
            file_ext = m.group(1).lower() if m else None
            if file_ext is not None:
                img_format = img_pillow.format.lower()
                if (img_format in ('png', 'gif',) and file_ext != img_format) \
                or (img_format == 'jpeg' and file_ext not in ('jpg', 'jpe', 'jpeg',)):
                    raise forms.ValidationError(
                        _("Image format doesn't match file extension. Please use another image.")
                    )
            
            # check dimensions in px
            if min_w is not None or min_h is not None:
                w, h = img_pillow.size
                err_txt = ''
                
                if min_w is not None and w < min_w:
                    err_txt += _('Image width must be minimum %(min_w)d px, this is %(w)d px.') \
                                    % { 'min_w': min_w, 'w': w }
                if min_h is not None and h < min_h:
                    err_txt += _(' Image height must be minimum %(min_h)d px, this is %(h)d px.') \
                                    % { 'min_h': min_h, 'h': h }
                if min_w is not None and min_h is not None \
                and round(min_w/min_h, 2) != round(w/h, 2):
                # rounding so we can make a small mistake in a couple of pixels,
                # the browser itself will correct that
                    ratio = TaskAdminForm._ratio_from_wh(min_w, min_h)
                    err_txt += _(' Wrong aspect ratio, %(w_ratio)d:%(h_ratio)d needed.') \
                                    % { 'w_ratio': ratio[0], 'h_ratio': ratio[1] }
                
                if err_txt:
                    raise forms.ValidationError(err_txt)
    
    def clean_run_at(self):
        run_at = self.cleaned_data.get('run_at')
        
        near_qs = Task.objects.filter(
            is_active=True,
            run_at__range=(run_at-timedelta(hours=2, minutes=59, seconds=59),
                           run_at+timedelta(hours=2, minutes=59, seconds=59))
        )
        if self.instance and self.instance.pk:  # exclude self from search
            near_qs = near_qs.exclude(id=self.instance.pk)
        if near_qs.exists():
            raise forms.ValidationError(
                _("You can't send pushes more than once per 6 hours. \
                  Please select another date and time.")
            )
        
        return run_at

    def clean_url(self):
        url = self.cleaned_data.get('url')
        
        if not settings.DEBUG:
            # replacing http with httpS - pushes will open relative
            # to service-worker domain, and it works (if non-local 127.0.0.1)
            # only in a secure context, so a push link must work over SSL
            url = re.sub(r'^http\:\/\/', 'https://', url, flags=re.I)
        
        # service-worker gets relative links, relative to it's own domain
        # so check that user provided link for the domain we control
        sites = Site.objects.all()
        if sites:
            our_domain = False
            for site in sites:
                if re.match(r'https?\:\/\/'+re.escape(site.domain)+r'\/', url, flags=re.I):
                    our_domain = True
                    break
            if not our_domain:
                raise forms.ValidationError(
                    _("Link must belong to one of your domains: %(domains)s.") \
                        % { 'domains': ', '.join([ s.domain for s in sites ]) }
                )
        
        # the link we push must work ))
        try:
            response_code = int( urllib.request.urlopen(url).getcode() )
            if response_code < 200 or response_code > 399:
                raise forms.ValidationError(
                    _("Erroneous HTTP %(response_code)d response from url.") \
                        % { 'response_code': response_code }
                )
        except IOError as e:
            raise forms.ValidationError(
                _("Got error while trying to access url: %(error_txt)s.") \
                    % { 'error_txt': str(e) }
            )
        
        return url
    


# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division, print_function

import string
import random
from unittest import skipUnless
from datetime import timedelta

from django.test import TestCase
from django.urls import reverse
from django.conf import settings
from django.utils import translation, timezone
from django.utils.six import StringIO
from django.contrib.sites.models import Site
#from django.contrib.auth import get_user_model
from django.core.management import call_command

from .settings import FCM_URL
from .models import DigestSubscription, Task


def __fake_letters(length=10):
    letters = string.ascii_letters + string.digits
    letters_len = len(letters)
    s = []
    for i in range(length):
        s.append(letters[ random.randint(0, letters_len-1) ])
    return ''.join(s)


def _new_subscription_obj_for_test(is_active=True):
    s = DigestSubscription()
    s.endpoint = FCM_URL + '/' + __fake_letters(30)
    s.timezone = settings.TIME_ZONE
    s.is_active = is_active
    if is_active:
        s.activated_at = timezone.now()
    s.save()
    return s


def _new_task_obj_for_test(is_active=True, run_at=None, started_at=None, done_at=None):
    t = Task()
    t.title = __fake_letters(10)
    t.message = __fake_letters(50)
    t.url = 'https://' + Site.objects.all()[0].domain + reverse('push_info_disable')
    t.is_active = is_active
    if run_at is not None:
        t.run_at = run_at
    t.save()
    t.started_at = started_at
    t.done_at = done_at
    # fake task as done
    if is_active and started_at is not None and done_at is not None:
        for tz_layout in t.timezonelayout_set.all():
            tz_layout.started_at = started_at
            tz_layout.done_at = done_at
            tz_layout.save()
    t.save()
    return t


class PushTests(TestCase):
    def test_push_on_off_view_working(self):
        response = self.client.get(reverse('push_on_off'))
        self.assertContains(response, 'js-push-button')

    def test_push_info_disable_view_working(self):
        with translation.override('en'):
            response = self.client.get(reverse('push_info_disable'))
        # 3 major content sections
        self.assertContains(response, 'Chrome')
        self.assertContains(response, 'Firefox')
        self.assertContains(response, 'Chrome for Android')
    
    def test_manifest_json_has_fcm_sender_id(self):
        response = self.client.get(reverse('push_manifest_json'))
        self.assertTrue( response.json()['gcm_sender_id'] )
    
    def test_view_saves_subscription_on_server(self):
        response = self.client.post(reverse('push_save'), {
            'endpoint': FCM_URL + '/hstwtdjsyTDSDGSU',
        })
        self.assertEqual(response.status_code, 200)
        self.assertTrue( int(response.json()['response']['id']) > 0 )
    
    def test_view_deactivates_subscription_on_server(self):
        obj = _new_subscription_obj_for_test(True)
        response = self.client.post(reverse('push_deactivate'), {
            'endpoint': obj.endpoint,
        })
        self.assertEqual(response.status_code, 200)
        self.assertEqual(int(response.json()['response']['id']), obj.pk)
        obj.refresh_from_db()
        self.assertFalse(obj.is_active)
    
    @skipUnless(settings.SITE_ID, "Can't generate push-task obj without sites framework.")
    def test_last_notification_view(self):
        obj = _new_task_obj_for_test(True,
                                     timezone.now()-timedelta(days=3),
                                     timezone.now(), timezone.now())
        response = self.client.get(reverse('push_last_notification'))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(obj.title, response.json()['notification']['title'])
        self.assertEqual(obj.message, response.json()['notification']['message'])
        self.assertTrue('url' in response.json()['notification'] and response.json()['notification']['url'])
        # last notification increases views counter for task
        obj.refresh_from_db()
        self.assertEqual(obj.views, 1)

    @skipUnless(settings.SITE_ID, "Can't generate push-task obj without sites framework.")
    def test_show_notification_view(self):
        obj = _new_task_obj_for_test(True,
                             timezone.now()-timedelta(days=3),
                             timezone.now(), timezone.now())
        response = self.client.get(reverse('push_show_notification', args=[obj.id,]))
        self.assertRedirects(response, obj.url+'?from=push')
        # show notification increases clicks counter
        obj.refresh_from_db()
        self.assertEqual(obj.clicks, 1)
    
    @skipUnless(settings.SITE_ID, "Can't generate push-task obj without sites framework.")
    def test_pushsend_management_command(self):
        subscr = _new_subscription_obj_for_test(True)
        task = _new_task_obj_for_test(True, timezone.now()-timedelta(days=3))
        out = StringIO()
        call_command('pushsend', stdout=out)
        subscr.refresh_from_db()
        # fake endpoint will have to give this subscription some error points
        self.assertTrue(subscr.errors > 0)
        task.refresh_from_db()
        self.assertTrue(task.done_at is not None)
        self.assertTrue(task.started_at is not None)
    
    # @skipUnless(settings.SITE_ID, "Can't generate push-task obj without sites framework.")
    # def test_task_admin_form(self):
    #     # enter /admin (superuser so we don't have to bother with permissions)
    #     user = get_user_model().objects.create_superuser(
    #         'username', 'admin@localhost', 'password123'
    #     )
    #     self.client.force_login(user)
    #     response = self.client.post(
    #         reverse('admin:push_task_add'), 
    #         {
    #             'title': 'push title',
    #             'message': 'push message',
    #             'url': '/',
    #             'run_at': timezone.now(),
    #         }
    #     )
    #     self.assertEqual(response.status_code, 200)
    #     # total 1 task should be in test DB now
    #     self.assertEqual(Task.objects.count(), 1)
    #     self.client.logout()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division, print_function

from django.utils.translation import ugettext as _

from commonstuff.filters import MyNullNotNullFilter


class IsDoneFilter(MyNullNotNullFilter):
    """
    Model field stores datetime when push task has been done, but in admin
    this filter reduces options to not sent (IS NULL) or already sent (NOT NULL).
    """

    title = _('already sent?')
    parameter_name = 'done_at'



# -*- coding: utf-8 -*-
import os
import setuptools


README = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


setuptools.setup(
    name='django_infopush',
    version='1.6.9',
    packages=setuptools.find_packages(),  # ['push'],
    include_package_data=True,
    license='MIT',
    description='Django reusable app, what allows to send web push.',
    long_description=README,
    long_description_content_type="text/x-rst",
    author='Yuriy Zemskov',
    author_email='zemskyura@gmail.com',
    url='https://github.com/kilgoretrout1985/django_infopush',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=[
        'Django>=1.11,<2.0',
        'django-commonstuff>=0.8.7',
        'pytz>=2018.3',
        'Pillow>=4.3.0',  # for dimensions on image upload
        'pywebpush>=1.7.0',  # payload encryption
    ]
)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division, print_function

from django.contrib import admin
from django.contrib.admin.views.main import ChangeList
from django.db.models import Sum
from django.utils.translation import ugettext_lazy as _, ungettext

from commonstuff.filters import MyDateListFilterNoFuture, MyDateListFilter

from .models import DigestSubscription, Task
from .forms import TaskAdminForm
from .filters import IsDoneFilter


class BaseSubscriptionAdmin(admin.ModelAdmin):
    list_display = ('endpoint_truncated', 'is_active', 'errors', 'created_at', \
                    'activated_at', 'deactivated_at',)
    list_filter = (
        'is_active',
        ('created_at', MyDateListFilterNoFuture),
        ('activated_at', MyDateListFilterNoFuture),
        ('deactivated_at', MyDateListFilterNoFuture),
    )
    search_fields = ('endpoint', 'ua', 'key', 'auth_secret', 'timezone')
    date_hierarchy = 'created_at'
    list_per_page = 20
    ordering = ('-created_at',)
    has_add_permission = lambda self, request: False
    readonly_fields = ['endpoint', 'key', 'auth_secret', 'errors', 'ua', \
                       'timezone', 'created_at', 'activated_at', \
                       'deactivated_at',]


class DigestSubscriptionAdmin(BaseSubscriptionAdmin):
    pass


class TaskChangeList(ChangeList):
    def get_results(self, request):
        super(TaskChangeList, self).get_results(request)
        totals = self.result_list.aggregate(
            total_views=Sum('views'),
            total_clicks=Sum('clicks'),
            total_closings=Sum('closings')
        )
        try:
            totals['avg_ctr'] = "%0.2f%%" % (totals['total_clicks'] * 100.0 / totals['total_views'])
        except (ZeroDivisionError, TypeError):
            totals['avg_ctr'] = 'n/a'
        try:
            totals['avg_closings'] = "%0.2f%%" % (totals['total_closings'] * 100.0 / totals['total_views'])
        except (ZeroDivisionError, TypeError):
            totals['avg_closings'] = 'n/a'
        self.totals = totals


class TaskAdmin(admin.ModelAdmin):
    form = TaskAdminForm
    list_display = ('title', 'message', 'is_active', 'has_image', 'has_image2',\
                    'views', 'clicks', 'ctr', 'closings_percent', 'run_at', \
                    'is_done', 'run_for')
    list_filter = (
        'is_active',
        IsDoneFilter,
        ('run_at', MyDateListFilter),
        ('done_at', MyDateListFilterNoFuture),
    )
    search_fields = ('title', 'message', 'url')
    date_hierarchy = 'run_at'
    list_per_page = 20
    list_max_show_all = 3000
    ordering = ('-run_at',)
    readonly_fields = ['started_at', 'done_at',]
    actions = ['activate_tasks', 'deactivate_tasks']
    
    def get_readonly_fields(self, request, obj=None):
        ro = self.readonly_fields[:]
        if obj and obj.started_at:
            ro.insert(0, 'run_at')  # maintain field order
        return ro
    
    def get_changelist(self, request):
        return TaskChangeList
    
    def _activate_deactivate_tasks(self, request, queryset, activate):
        counter = 0
        for task in queryset:
            if task.is_active != activate:
                task.is_active = activate
                task.save()
                counter += 1
        if counter:
            if activate:
                msg = ungettext(
                    'Successfully activated %(count)d task.',
                    'Successfully activated %(count)d tasks.',
                    counter
                ) % {'count': counter}
            else:
                msg = ungettext(
                    'Successfully deactivated %(count)d task.',
                    'Successfully deactivated %(count)d tasks.',
                    counter
                ) % {'count': counter}
            self.message_user(request, msg)
    
    def deactivate_tasks(self, request, queryset):
        self._activate_deactivate_tasks(request, queryset, False)
    deactivate_tasks.short_description = _('Deactivate selected tasks')

    def activate_tasks(self, request, queryset):
        self._activate_deactivate_tasks(request, queryset, True)
    activate_tasks.short_description = _('Activate selected tasks')


admin.site.register(Task, TaskAdmin)
admin.site.register(DigestSubscription, DigestSubscriptionAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import, division, print_function

from django.conf.urls import url
from . import views


urlpatterns = [
    # subscriptions
    url(r'^on[\-\_]off/$', views.on_off, name='push_on_off'),
    url(r'^info[\-\_]disable/$', views.info_disable, name='push_info_disable'),
    url(r'^manifest\.json$', views.manifest_json, name='push_manifest_json'),
    url(r'^save/$', views.save, name='push_save'),
    url(r'^deactivate/$', views.deactivate, name='push_deactivate'),
    
    # tasks
    url(r'^last_notification/$', views.last_notification, name='push_last_notification'),
    url(r'^show_notification/(\d+)/$', views.show_notification, name='push_show_notification'),
    url(r'^notification_plus_one/(views|closings)/(\d+)/$',
        views.notification_plus_one, name='push_notification_plus_one'),
]
